/*****************************************************************************/
/*  config_file.c - config file routines                                     */
/*  Copyright (C) 1998-1999 Brian Masney <masneyb@newwave.net>               */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either version 2 of the License, or        */
/*  (at your option) any later version.                                      */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA      */
/*****************************************************************************/

#include "ftp.h"

extern struct conn_categories *hostcat;
extern GHashTable *bookmarks_htable;
extern struct pix_ext *registered_exts;
extern char firewall_host[MAXSTR], firewall_username[MAXSTR], firewall_password[MAXSTR],
            view_program[MAXSTR], edit_program[MAXSTR], emailaddr[MAXSTR];
extern int use_cache, use_firewall, firewall_port, start_file_transfers, do_one_transfer_at_a_time,
           listbox_local_width, listbox_remote_width, listbox_file_height,
           transfer_height, log_height, save_geometry, passive_transfer,
           listbox_filename_width, sort_dirs_first, use_default_dl_types, 
           confirm_delete, refresh_files, preserve_attribs, listbox_size_width,
           listbox_user_width, listbox_group_width, listbox_date_width,
           listbox_attribs_width, use_same_fd, timeout, retries, sleep_time;
extern struct ftp_window_data window2;
extern GtkWidget *firewall_btn;

static int copyfile (char *source, char *dest);
static gint hash_compare (gconstpointer path1, gconstpointer path2);
static guint hash_function (gconstpointer key);

struct config_vars {
   char key[30], /* Our config file variable name */
        description[35], /* How this field will show up in the options dialog */
        type; /* Type Character or Integer */
   void *var; /* Pointer to our variable */
   int len, /* Length of this field. Used only for char strings */
       general_option : 1; /* Will this go into the general settings tab */
   char comment[MAXSTR]; /* Comment to write out to the config file */
   GtkWidget *widget;
}; 

struct proxy_type_tag {
   char key[30];
   int show_host_port,
       show_user_pass;
};

static void proxy_toggle(GtkList *list, GtkWidget *child, gpointer data);
static void apply_changes (GtkWidget *widget, gpointer data);
static int proxy_num;

struct config_vars config_file_vars[] = {
/* You MUST keep the firewall stuff up here as the first entries and in the 
   same order. The options dialog relies on it being here */
   {"usefirewall", 		"Connect via FTP proxy", 	'I', &use_firewall, 		0, 			0, "# (*) Connect to the ftp server via a FTP Proxy.\n# 0 = Disable firewall support\n# 1 = Login firewall via SITE command\n# 2 = Login firewall via USER/PASS\n# 3 = Login via USER user@host:port", NULL},
   {"firewall_host", 		"Proxy hostname:", 		'C', firewall_host, 		sizeof(firewall_host), 	0, "# (*) Firewall hostname", NULL},
   {"firewall_port", 		"Proxy port:", 			'I', &firewall_port, 		10,			0, "# (*) Port to connect to on the firewall", NULL},
   {"firewall_username", 	"Proxy username:", 		'C', firewall_username,		sizeof(firewall_username), 0, "# (*) Your firewall username", NULL},
   {"firewall_password", 	"Proxy password:", 		'C', firewall_password, 	sizeof(firewall_password), 0, "# (*) Your firewall password", NULL},
   {"email", 			"Email address:", 		'C', emailaddr, 		sizeof(emailaddr), 	1, "# Enter your email address here", NULL},
   {"view_program", 		"View program:", 		'C', view_program, 		sizeof(view_program), 	1, "# The default program used to view files. If this is blank, the internal\n# file viewer will be used", NULL},
   {"edit_program", 		"Edit program:", 		'C', edit_program, 		sizeof(edit_program), 	1, "# The default program used to edit files.", NULL},
   {"connect_timeout",		"Connect timeout:",		'I', &timeout,			10,			1, "# The max timeout for the connection", NULL},
   {"retries",			"Connect retries:",		'I', &retries,			10,			1, "# The number of auto-retries to do", NULL},
   {"sleep_time",		"Retry sleep time:",		'I', &sleep_time,		10,			1, "# The number of seconds to wait between retries", NULL},
   {"confirm_delete",		"Confirm delete",		'I', &confirm_delete,		0,			1, "# Confirm when deleting files", NULL},
   {"one_transfer", 		"Do one transfer at a time",	'I', &do_one_transfer_at_a_time,0, 			1, "# Do only one transfer at a time?", NULL},
   {"passive_transfer",		"Passive file transfers", 	'I', &passive_transfer,		0, 			1, "# Send PASV command or PORT command for data transfers", NULL},
   {"preserve_permissions",	"Preserve permissions",		'I', &preserve_attribs,		0,			1, "# Save the permissions of the transfered files", NULL},
   {"refresh_files",		"Refresh after each file transfer",	'I', &refresh_files,		0,			1, "# Refresh the listbox after each file is transfered", NULL},
   {"save_geometry", 		"Save geometry", 		'I', &save_geometry, 		0, 			1, "# Save the size of each widget for next startup", NULL},
   {"sort_dirs_first", 		"Sort directories first", 	'I', &sort_dirs_first, 		0, 			1, "# Put the directories first then the files", NULL},
   {"start_transfers",		"Start file transfers", 	'I', &start_file_transfers, 	0, 			1, "# Automatically start the file transfers when they get queued?", NULL},
   {"usecache", 		"Use cache", 			'I', &use_cache, 		0, 			1, "# Do you want to use the cache?", NULL},
   {"use_same_fd",		"Use same connection",		'I', &use_same_fd,		0,			1, "# Use same connection for the transfers. Will only log in once", NULL},
   {"use_default_dl_types", 	"", 				'I', &use_default_dl_types, 	0, 			0, "# (*) If this is set, and there is a ext= line below for the file extension,\n# it will download the file as specified below", NULL},
   {"listbox_local_width",	"", 				'I', &listbox_local_width, 	0, 			0, "# The default width of the local files listbox", NULL},
   {"listbox_remote_width", 	"", 				'I', &listbox_remote_width, 	0, 			0, "# The default width of the remote files listbox", NULL},
   {"listbox_file_height", 	"", 				'I', &listbox_file_height, 	0, 			0, "# The default height of the local/remote files listboxes", NULL},
   {"transfer_height", 		"", 				'I', &transfer_height, 		0, 			0, "# The default height of the transfer listbox", NULL},
   {"log_height", 		"", 				'I', &log_height, 		0, 			0, "# The default height of the logging window", NULL},
   {"filename_width", 		"", 				'I', &listbox_filename_width, 	0, 			0, "# (*) The width of the filename column in the file listboxes", NULL},
   {"size_width", 		"", 				'I', &listbox_size_width, 	0, 			0, "# (*) The width of the size column in the file listboxes", NULL},
   {"user_width", 		"", 				'I', &listbox_user_width, 	0, 			0, "# (*) The width of the user column in the file listboxes", NULL},
   {"group_width", 		"", 				'I', &listbox_group_width, 	0, 			0, "# (*) The width of the group column in the file listboxes", NULL},
   {"date_width", 		"", 				'I', &listbox_date_width, 	0, 			0, "# (*) The width of the date column in the file listboxes", NULL},
   {"attribs_width", 		"", 				'I', &listbox_attribs_width, 	0, 			0, "# (*) The width of the attribs column in the file listboxes", NULL},
   {"", "", 0, NULL}};

struct proxy_type_tag proxy_type[] = {
   {"none", 		0, 0},
   {"SITE command", 	1, 1},
   {"user@host", 	1, 1},
   {"user@host:port",	1, 0},
   {"AUTHENTICATE",	1, 1},
   {"user@host port",	1, 0},
   {"user@host NOAUTH", 1, 0},
   {"HTTP Proxy",	1, 0},
   {"", 0, 0}};

#define HTTP_PROXY_NUM	7

void options_dialog(gpointer data) {
   GtkWidget *tempwid, *dialog, *notebook, *box, *intbox, *chartbl, *table, *label;
   GList *proxy_list;
   char tempstr[30];
   int num, tbl_len;

   proxy_list = NULL;   
   dialog = gtk_dialog_new ();
   gtk_window_set_title (GTK_WINDOW (dialog), "Options");
   gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 10);
   gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (dialog)->action_area), 5);
   gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 2);
   gtk_box_set_homogeneous (GTK_BOX (GTK_DIALOG (dialog)->action_area), TRUE);
   gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
   gtk_signal_connect_object (GTK_OBJECT (dialog), "delete_event", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT( dialog));
   gtk_signal_connect_object (GTK_OBJECT (dialog), "destroy", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog));
   
   notebook = gtk_notebook_new ();
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), notebook, TRUE, TRUE, TRUE);
   gtk_widget_show (notebook);
   
   box = gtk_vbox_new (FALSE, 0);
   gtk_container_border_width (GTK_CONTAINER (box), 10);
   gtk_widget_show (box);

   tempwid = gtk_label_new ("General");
   gtk_widget_show (tempwid);
   gtk_notebook_append_page (GTK_NOTEBOOK (notebook), box, tempwid);   

   tbl_len = 1;
   chartbl = gtk_table_new (tbl_len, 2, FALSE);
   gtk_table_set_row_spacings (GTK_TABLE (chartbl), 5);
   gtk_table_set_col_spacings (GTK_TABLE (chartbl), 5);
   gtk_box_pack_start (GTK_BOX (box), chartbl, FALSE, FALSE, FALSE);
   gtk_widget_show (chartbl);
   
   intbox = gtk_vbox_new (FALSE, 0);
   gtk_box_pack_start (GTK_BOX (box), intbox, TRUE, TRUE, TRUE);
   gtk_widget_show (intbox);

   num = 0;
   while (config_file_vars[num].var != NULL) {
      if (config_file_vars[num].general_option) {
         if (config_file_vars[num].type == 'I' && config_file_vars[num].len == 0) {
            tempwid = gtk_check_button_new_with_label (config_file_vars[num].description);
            gtk_box_pack_start (GTK_BOX (intbox), tempwid, FALSE, FALSE, FALSE);
            gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (tempwid), *(int *) config_file_vars[num].var);
            gtk_widget_show (tempwid);
            config_file_vars[num].widget = tempwid;
         }
         else {
            tbl_len++;
            gtk_table_resize (GTK_TABLE (chartbl), tbl_len, 2);
            
            tempwid = gtk_label_new (config_file_vars[num].description);
            gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
            gtk_table_attach_defaults (GTK_TABLE (chartbl), tempwid, 0, 1, tbl_len-1, tbl_len);
            gtk_widget_show (tempwid);

            tempwid = gtk_entry_new_with_max_length (config_file_vars[num].len);
            gtk_table_attach_defaults (GTK_TABLE (chartbl), tempwid, 1, 2, tbl_len-1, tbl_len);
            if (config_file_vars[num].type == 'I') {
               g_snprintf (tempstr, sizeof (tempstr), "%d", *(int *) config_file_vars[num].var);
               gtk_entry_set_text (GTK_ENTRY (tempwid), tempstr);
            }
            else {
               gtk_entry_set_text (GTK_ENTRY (tempwid), config_file_vars[num].var);
            }
            gtk_widget_show (tempwid);
            config_file_vars[num].widget = tempwid;
         }
      }
      num++;
   }

   box = gtk_vbox_new (FALSE, 5);
   gtk_container_border_width (GTK_CONTAINER (box), 10);
   gtk_widget_show (box);

   tempwid = gtk_label_new ("Proxy Server");
   gtk_widget_show (tempwid);
   gtk_notebook_append_page (GTK_NOTEBOOK (notebook), box, tempwid);   

   table = gtk_table_new (5, 2, FALSE);
   gtk_table_set_row_spacings (GTK_TABLE (table), 5);
   gtk_table_set_col_spacings (GTK_TABLE (table), 5);
   gtk_box_pack_start (GTK_BOX (box), table, FALSE, FALSE, FALSE);
   gtk_widget_show (table);

   tempwid = gtk_label_new ("Proxy server type");
   gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 0, 1);
   gtk_widget_show (tempwid);

   tempwid = gtk_label_new (config_file_vars[1].description);
   gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 1, 2);
   gtk_widget_show (tempwid);

   tempwid = gtk_entry_new_with_max_length (config_file_vars[1].len);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 1, 2, 1, 2);
   gtk_entry_set_text (GTK_ENTRY (tempwid), config_file_vars[1].var);
   gtk_widget_show (tempwid);
   config_file_vars[1].widget = tempwid;

   tempwid = gtk_label_new (config_file_vars[2].description);
   gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 2, 3);
   gtk_widget_show (tempwid);

   tempwid = gtk_entry_new_with_max_length (config_file_vars[2].len);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 1, 2, 2, 3);
   g_snprintf (tempstr, sizeof (tempstr), "%d", *(int *) config_file_vars[2].var);
   gtk_entry_set_text (GTK_ENTRY (tempwid), tempstr);
   gtk_widget_show (tempwid);
   config_file_vars[2].widget = tempwid;

   tempwid = gtk_label_new (config_file_vars[3].description);
   gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 3, 4);
   gtk_widget_show (tempwid);

   tempwid = gtk_entry_new_with_max_length (config_file_vars[3].len);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 1, 2, 3, 4);
   gtk_entry_set_text (GTK_ENTRY (tempwid), config_file_vars[3].var);
   gtk_widget_show (tempwid);
   config_file_vars[3].widget = tempwid;

   tempwid = gtk_label_new (config_file_vars[4].description);
   gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 4, 5);
   gtk_widget_show (tempwid);

   tempwid = gtk_entry_new_with_max_length (config_file_vars[4].len);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 1, 2, 4, 5);
   gtk_entry_set_text (GTK_ENTRY (tempwid), config_file_vars[4].var);
   gtk_entry_set_visibility (GTK_ENTRY (tempwid), FALSE);
   gtk_widget_show (tempwid);
   config_file_vars[4].widget = tempwid;

   tempwid = gtk_combo_new ();
   gtk_signal_connect (GTK_OBJECT (GTK_COMBO (tempwid)->list), "select_child", GTK_SIGNAL_FUNC (proxy_toggle), NULL);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 1, 2, 0, 1);
   gtk_widget_show (tempwid);
   config_file_vars[0].widget = tempwid;

   num = 0;
   while (*proxy_type[num].key != '\0') {
      label = gtk_list_item_new_with_label (proxy_type[num].key);
      gtk_widget_show (label);
      proxy_list = g_list_append (proxy_list, label);
      num++;
   }
   gtk_list_prepend_items (GTK_LIST (GTK_COMBO(tempwid)->list), proxy_list);
   proxy_num = *(int *) config_file_vars[0].var;
   if (proxy_num == HTTP_PROXY) proxy_num = HTTP_PROXY_NUM;
   gtk_list_select_item (GTK_LIST (GTK_COMBO (tempwid)->list), proxy_num);

   tempwid = gtk_button_new_with_label ("OK");
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, TRUE);
   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (apply_changes), (gpointer) NULL);
   gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog));
   gtk_widget_show (tempwid);

   tempwid = gtk_button_new_with_label ("  Cancel  ");
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, TRUE);
   gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog));
   gtk_widget_show (tempwid);

   tempwid = gtk_button_new_with_label ("Apply");
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, TRUE);
   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (apply_changes), (gpointer) NULL);
   gtk_widget_grab_default (tempwid);
   gtk_widget_show (tempwid);

   gtk_widget_show (dialog);
}
/*****************************************************************************/
static void proxy_toggle(GtkList *list, GtkWidget *child, gpointer data) {
   gint pos;
   
   pos = gtk_list_child_position (list, child);
   proxy_num = pos;
   gtk_widget_set_sensitive (config_file_vars[1].widget, proxy_type[proxy_num].show_host_port);
   gtk_widget_set_sensitive (config_file_vars[2].widget, proxy_type[proxy_num].show_host_port);
   gtk_widget_set_sensitive (config_file_vars[3].widget, proxy_type[proxy_num].show_user_pass);
   gtk_widget_set_sensitive (config_file_vars[4].widget, proxy_type[proxy_num].show_user_pass);
}
/*****************************************************************************/
static void apply_changes (GtkWidget *widget, gpointer data) {
   char *tempstr;
   int num;
   
   if (proxy_num == HTTP_PROXY_NUM) proxy_num = HTTP_PROXY;
   *(int *) config_file_vars[0].var = proxy_num;
   num = 1;
   while (config_file_vars[num].var != NULL) {
      if (config_file_vars[num].widget != NULL) {
         if(config_file_vars[num].type == 'I' && config_file_vars[num].len == 0) {
            *(int *) config_file_vars[num].var = GTK_TOGGLE_BUTTON (config_file_vars[num].widget)->active;
         }
         else {
            tempstr = gtk_entry_get_text (GTK_ENTRY (config_file_vars[num].widget));
            if (config_file_vars[num].type == 'I') {
               *(int *) config_file_vars[num].var = strtol (tempstr, (char **) NULL, 10);
            }
            else {
               strncpy ((char *) config_file_vars[num].var, tempstr, config_file_vars[num].len);
               ((char *) config_file_vars[num].var)[config_file_vars[num].len-1] = '\0';
            }
         }
      }
      num++;
   }
   window2.hdata->ftpdata->proxy_type = proxy_num;
   gftp_set_proxy_hostname (window2.hdata->ftpdata, firewall_host);
   gftp_set_proxy_username (window2.hdata->ftpdata, firewall_username);
   gftp_set_proxy_password (window2.hdata->ftpdata, firewall_password);
   gftp_set_proxy_port (window2.hdata->ftpdata, firewall_port);
   GFTP_SET_TRANSFER_TYPE (window2.hdata->ftpdata,
   	passive_transfer ? gftp_transfer_passive : gftp_transfer_active);
   gtk_widget_set_sensitive (firewall_btn, proxy_num > 0);
   write_config_file ();
}
/*****************************************************************************/
void read_config_file (void) {
   struct conn_categories *newentry, *preventry, *folderentry, *endentry;
   struct pix_ext *tempext;
   char tempstr[MAXSTR], temp1str[MAXSTR], *curpos, tempchar;
   FILE *conffile;
   int line=0, pos;
   
   hostcat = g_malloc0 (sizeof (struct conn_categories));
   hostcat->isfolder = 1;
   hostcat->path = g_malloc (1);
   *hostcat->path = '\0';
   bookmarks_htable = g_hash_table_new (hash_function, hash_compare);

   registered_exts = NULL;
   preventry = NULL;
   if (!expand_path (CONFIG_FILE, tempstr, sizeof (tempstr))) {
      printf ("gFTP Error: Bad config file name %s\n", CONFIG_FILE);
      exit (0);
   }
   if (access (tempstr, F_OK) == -1) {
      expand_path (BASE_CONF_DIR, temp1str, sizeof (temp1str));
      if (access (temp1str, F_OK) == -1) {
         if (mkdir (temp1str, 488) != 0) {
            printf ("gFTP Error: Could not make directory %s: %s\n", temp1str, g_strerror(errno));
            exit (-1);
         }
      }
      g_snprintf (temp1str, sizeof(temp1str), "%s/gftprc", SHARE_DIR);
      temp1str[sizeof (temp1str) - 1] = '\0';
      if (access (temp1str, F_OK) == -1) {
         printf ("gFTP Error: Cannot find master config file %s\n", temp1str);
         printf ("Did you do a make install?\n");
         exit (-1);
      }
      copyfile (temp1str, tempstr);
   }
   chmod (tempstr, S_IRUSR | S_IWUSR);
   conffile = fopen (tempstr, "r");
   if (conffile == NULL) {
      printf ("gFTP Error: Cannot open config file %s: %s\n", CONFIG_FILE, g_strerror (errno));
      exit (0);
   }
   while (fgets (tempstr, sizeof (tempstr), conffile)) {
      if (tempstr[strlen (tempstr) - 1] == '\n') {
         tempstr[strlen (tempstr) - 1] = '\0';
         if (tempstr[strlen (tempstr) - 2] == '\r') tempstr[strlen (tempstr) - 2] = '\0';
      }
      line++;
      if (strncmp (tempstr, "host=", 5) == 0) {
         curpos = tempstr+5;
         newentry = g_malloc0 (sizeof (struct conn_categories));
         newentry->isfolder = 0;
         parse_args (curpos, 7, line,
            &newentry->path, 0,
            &newentry->hostname, 0,
            &newentry->port, 0,
            &newentry->dir, 0,
            &newentry->user, 0,
            &newentry->pass, 0,
            &newentry->firewall, 0);

         /* We have to create the folders. For example, if we have 
            Debian Sites/Debian, we have to create a Debian Sites entry */
         preventry = hostcat;
         if (preventry->children != NULL) { 
            endentry = preventry->children;
            while (endentry->next != NULL) endentry = endentry->next;
         }
         else endentry = NULL;
         curpos = newentry->path;
         while ((curpos = strchr (curpos, '/')) != NULL) {
            tempchar = *curpos;
            *curpos = '\0';
            /* See if we already made this folder */
            if ((folderentry = (struct conn_categories *) 
            	g_hash_table_lookup (bookmarks_htable, newentry->path)) == NULL) {

               /* Allocate the individual folder. We have to do this for the edit 
                  bookmarks feature */
               folderentry = g_malloc0 (sizeof (struct conn_categories));
               folderentry->path = g_malloc (strlen (newentry->path) + 1);
               strcpy (folderentry->path, newentry->path);
               folderentry->prev = preventry;
               folderentry->isfolder = 1;
               g_hash_table_insert (bookmarks_htable, folderentry->path, folderentry);
               if (preventry->children == NULL) preventry->children = folderentry;
               else endentry->next = folderentry;
               preventry = folderentry;
               endentry = NULL;
            }
            else {
               preventry = folderentry;
               if (preventry->children != NULL) {
                  endentry = preventry->children;
                  while (endentry->next != NULL) endentry = endentry->next;
               }
               else endentry = NULL;
            }
            *curpos = tempchar;
            curpos++;
         }

         /* Get the parent node */
         if ((curpos = strrchr (newentry->path, '/')) == NULL) preventry = hostcat;
         else {
            tempchar = *curpos;
            *curpos = '\0';
            preventry = (struct conn_categories *) 
            		g_hash_table_lookup (bookmarks_htable, newentry->path);
            *curpos = tempchar;
         }
         
         if (preventry->children != NULL) {
            endentry = preventry->children;
            while (endentry->next != NULL) endentry = endentry->next;
            endentry->next = newentry;
         }
         else preventry->children = newentry;
         newentry->prev = preventry;
         g_hash_table_insert (bookmarks_htable, newentry->path, newentry);
      }
      else if (strncmp(tempstr, "ext=", 4) == 0) {
         curpos = tempstr+4;
         tempext = g_malloc (sizeof (struct pix_ext));
         parse_args (curpos, 4, line, 
            &tempext->ext, 0,
            &tempext->filename, 0,
            &tempext->ascii_binary, 0,
            &tempext->view_program, 0);
         g_snprintf (tempstr, sizeof (tempstr), "%s/%s", BASE_CONF_DIR, tempext->filename);
         tempstr[sizeof (tempstr) - 1] = '\0';
         if (!tempext->filename || !expand_path (tempstr, temp1str, sizeof (temp1str))) {
            printf ("gFTP Error: Bad file name %s in config file on line %d\n", tempext->filename, line);
            exit (0);
         }
         if (*tempext->filename != '\0' && access (temp1str, F_OK) != 0) {
            g_snprintf (tempstr, sizeof (tempstr), "%s/%s", SHARE_DIR, tempext->filename);
            tempstr[sizeof (tempstr) - 1] = '\0';
            if (!expand_path (tempstr, temp1str, sizeof (temp1str))) {
               g_snprintf (tempstr, sizeof (tempstr), "/usr/share/icons/%s", tempext->filename);
               tempstr[sizeof (tempstr) - 1] = '\0';
               if (!expand_path (tempstr, temp1str, sizeof (temp1str))) {
                  printf ("gFTP Error: Bad file name %s in config file on line %d\n", tempext->filename, line);
                  exit (0);
               }
            }
            if (access (temp1str, F_OK) != 0) {
               printf ("gFTP Error: Error on line %d: %s doesn't exist in %s or %s\n", line, tempext->filename, SHARE_DIR, BASE_CONF_DIR);
               exit (0);
            }
         }
         tempext->stlen = strlen (tempext->ext);
         tempext->next = registered_exts;
         registered_exts = tempext;
      }         
      else if (*tempstr != '#' && *tempstr != '\0') {
         pos = 0;
         while (config_file_vars[pos].var != NULL) {
            if (strncmp (config_file_vars[pos].key, tempstr, strlen (config_file_vars[pos].key)) == 0) {
               curpos = tempstr + strlen (config_file_vars[pos].key) + 1;
               if (config_file_vars[pos].type == 'C') {
                  strncpy (config_file_vars[pos].var, curpos, config_file_vars[pos].len);
                  ((char *) config_file_vars[pos].var)[config_file_vars[pos].len-1] = '\0';
                  break;
               }
               else {
                  *(int *) config_file_vars[pos].var = strtol (curpos, (char **) NULL, 10);
                  break;
               }
            }
            pos++;
         }

         if (config_file_vars[pos].var == NULL) {
            printf ("gFTP Warning: Skipping line %d in config file: %s\n", line, tempstr);
         }
      }
   }
   return;
}
/*****************************************************************************/
void write_config_file (void) {
   const char *hdr = "# Config file for gFTP\n# Copyright (C) 1998-1999 Brian Masney <masneyb@newwave.net>\n# Warning: Any comments that you add to this file WILL be overwritten\n# If a entry has a (*) in it's comment, you can't change it inside gFTP\n\n";
   const char *hosthdr = "# host=menu path:hostname:port:start dir:username:password:firewall\n# If you set the password to @EMAIL@, gFTP will automatically change that to\n# your email address.\n# If you set the firewall argument to 1, gFTP will try to connect to your\n# FTP proxy if you have one. It is usually best to leave this set at 1 because\n# gFTP won't use it if you don't have a FTP proxy\n";
   const char *exthdr = "\n# ext=file extenstion:XPM file:Ascii or Binary (A or B):viewer program\n# Note: All arguments except the file extension are optional\n";
   struct conn_categories *tempentry;
   struct pix_ext *tempext;
   char tempstr[MAXSTR];
   FILE *conffile;
   int pos;

   if (!expand_path (CONFIG_FILE, tempstr, sizeof (tempstr))) {
      printf ("gFTP Error: Bad config file name %s\n", CONFIG_FILE);
      exit (0);
   }
   conffile = fopen (tempstr, "w+");
   if (conffile == NULL) {
      printf ("gFTP Error: Cannot open config file %s: %s\n", CONFIG_FILE, g_strerror (errno));
      exit (0);
   }
   fwrite (hdr, 1, strlen (hdr), conffile);

   pos = 0;
   while (config_file_vars[pos].var != NULL) {
      fwrite (config_file_vars[pos].comment, 1, strlen (config_file_vars[pos].comment), conffile);
      fwrite ("\n", 1, 1, conffile);
      if (config_file_vars[pos].type == 'C') {
         g_snprintf (tempstr, sizeof (tempstr), "%s=%s\n\n", 
            config_file_vars[pos].key, (char *) config_file_vars[pos].var);
      }
      else {
         g_snprintf (tempstr, sizeof(tempstr), "%s=%d\n\n", 
            config_file_vars[pos].key, *(int *) config_file_vars[pos].var);
      }
      tempstr[sizeof (tempstr) - 1] = '\0';
      fwrite (tempstr, 1, strlen (tempstr), conffile);
      pos++;
   }

   fwrite (hosthdr, 1, strlen (hosthdr), conffile);
   tempentry = hostcat->children;
   while (tempentry != NULL) {
      if (tempentry->children != NULL) {
         tempentry = tempentry->children;
         continue;
      }
      g_snprintf (tempstr, sizeof (tempstr), "host=%s:%s:%s:%s:%s:%s:%c\n", 
      	tempentry->path, tempentry->hostname, tempentry->port, tempentry->dir, 
      	tempentry->user, tempentry->pass, 
      	tempentry->firewall[0] == '\0' ? '1' : tempentry->firewall[0]);
      tempstr[sizeof (tempstr) - 1] = '\0';
      fwrite (tempstr, 1, strlen (tempstr), conffile);
      if (tempentry->next == NULL) {
         tempentry = tempentry->prev;
         while (tempentry->next == NULL && tempentry->prev != NULL) tempentry = tempentry->prev;
         tempentry = tempentry->next;
      }
      else tempentry = tempentry->next;
   }
   
   fwrite (exthdr, 1, strlen (exthdr), conffile);
   tempext = registered_exts;
   while (tempext != NULL) {
      g_snprintf (tempstr, sizeof (tempstr), "ext=%s:%s:%c:%s\n", 
      	tempext->ext, tempext->filename, 
      	*tempext->ascii_binary == '\0' ? ' ' : *tempext->ascii_binary, 
      	tempext->view_program);
      tempstr[sizeof (tempstr) - 1] = '\0';
      fwrite (tempstr, 1, strlen (tempstr), conffile);
      tempext = tempext->next;
   }
   fclose (conffile);
}
/*****************************************************************************/
static int copyfile (char *source, char *dest) {
   FILE *srcfd, *destfd;
   char buf[8192];
   size_t n;
   
   if ((srcfd = fopen (source, "rb")) == NULL) {
      return (0); 
   }
   if ((destfd = fopen (dest, "wb")) == NULL) {
      fclose (srcfd);
      return (0);
   }

   while ((n = fread (buf, 1, sizeof (buf), srcfd)) > 0) {
      fwrite (buf, 1, n, destfd);
   }

   fclose (srcfd);
   fclose (destfd);
   return (1);
}
/*****************************************************************************/
int parse_args (char *str, int numargs, int lineno, char **first, ...) {
   char *curpos, *endpos, **dest, tempchar;
   size_t len, newlen;
   va_list argp;
   int ret;
   
   ret = 1;
   va_start (argp, first);
   curpos = str;
   dest = first;
   len = va_arg (argp, size_t);
   memset (*dest, 0, len);
   while (numargs > 1) {
      endpos = strchr (curpos, ':');
      if (endpos == NULL) {
         printf ("gFTP Warning: Line %d doesn't have enough arguments\n", lineno);
         ret = 0;
         break;
      }
      if (len == 0) {
         *dest = g_malloc (endpos - curpos + 1);
         tempchar = *endpos;
         *endpos = '\0';
         strcpy (*dest, curpos);
         *endpos = tempchar;
      }
      else {
         newlen = endpos - curpos + 1 > len ? len : endpos - curpos + 1;
         strncpy (*dest, curpos, newlen);
         dest[0][newlen - 1] = '\0';
      }
      curpos = endpos + 1;
      dest = va_arg (argp, char **);
      len = va_arg (argp, size_t);
      memset (*dest, 0, len);
      numargs--;
   }
   if (len == 0) {
      *dest = g_malloc (strlen (curpos) + 1);
      strcpy (*dest, curpos);
   }
   else {
      newlen = strlen (curpos) + 1 > len ? len : strlen (curpos) + 1;
      strncpy(*dest, curpos, newlen);
      dest[0][newlen-1] = '\0';
   }

   numargs--;
   while (numargs > 1) {
      dest = va_arg (argp, char **);
      len = va_arg (argp, size_t);
      memset (*dest, 0, len);
      numargs--;
   }
   va_end (argp);
   return (1);
}
/*****************************************************************************/
GHashTable *build_hash_table (struct conn_categories *entry) {
   struct conn_categories *tempentry;
   GHashTable *htable;
   
   htable = g_hash_table_new (hash_function, hash_compare);
   tempentry = entry;
   while (tempentry != NULL) {
      g_hash_table_insert (htable, tempentry->path, tempentry);
      if (tempentry->children != NULL) {
         tempentry = tempentry->children;
         continue;
      }
      else {
         if (tempentry->next == NULL) {
            tempentry = tempentry->prev;
            while (tempentry->next == NULL && tempentry->prev != NULL) tempentry = tempentry->prev;
            tempentry = tempentry->next;
         }
         else tempentry = tempentry->next;
      }
   }
   return (htable);
}
/*****************************************************************************/
static gint hash_compare (gconstpointer path1, gconstpointer path2) {
   return (!strcmp ((char *) path1, (char *) path2));
}
/*****************************************************************************/
static guint hash_function (gconstpointer key) {
   return (((char *) key)[0] + ((char *) key)[1]);
}
/*****************************************************************************/
