/*****************************************************************************/
/*  view_dialog.c - view dialog box and ftp routines                         */
/*  Copyright (C) 1998-1999 Brian Masney <masneyb@newwave.net>               */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either version 2 of the License, or        */
/*  (at your option) any later version.                                      */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA      */
/*****************************************************************************/

#include "ftp.h"

extern struct pix_ext *registered_exts;
extern pthread_mutex_t sigchld_mutex;
extern struct viewedit_data *viewedit_processes;
extern char view_program[MAXSTR], edit_program[MAXSTR];

static void delete_file (GtkWidget *widget, char *fle);
static struct viewedit_data *fork_process (char *proc, char *filename, char *remote_filename, int viewedit, int del_file);
static struct ftp_window_data *wdata;
static struct ftp_file_data *curfle;

void view_dialog (gpointer data) {
   struct ftp_transfer_data *newtdata;
   char tempstr[MAXSTR];
   
   wdata = (struct ftp_window_data *) data;
   if (!check_status ("View", wdata, 1, 1, 0)) return;

   if ((curfle = get_next_selected_filename (wdata->hdata->files)) == NULL) {
      ftp_log (gftp_logging_misc, NULL, "Internal gFTP Error: Could not find a selected file. This is probably a bug. Please email masneyb@newwave.net about it\n");
      return;
   }

   if (curfle->flags & FILE_ISDIR) {
      ftp_log (gftp_logging_misc, NULL, "View: %s is a directory. Cannot view it.\n", curfle->file);
      return;
   }
   
   if (wdata->local) view_file (curfle->file, 1, 0, 1, NULL);
   else {
      make_temp_filename (tempstr, strchr (curfle->file, '.'), sizeof (tempstr));
      newtdata = transfer_one_file (tempstr, curfle->file, 1);
      newtdata->hdata->files->flags |= FILE_TRANS_DONE_VIEW;
      add_file_transfer (newtdata);
   }
}
/*****************************************************************************/
void edit_dialog (gpointer data) {
   struct ftp_transfer_data *newtdata;
   char tempstr[MAXSTR];
   
   wdata = (struct ftp_window_data *) data;
   if (!check_status ("Edit", wdata, 1, 1, 0)) return;

   if (*edit_program == '\0') {
      ftp_log (gftp_logging_misc, NULL, "Edit: You must specify an editor in the options dialog\n");
      return;
   }

   if ((curfle = get_next_selected_filename(wdata->hdata->files)) == NULL) {
      ftp_log (gftp_logging_misc, NULL, "Internal gFTP Error: Could not find a selected file. This is probably a bug. Please email masneyb@newwave.net about it\n");
      return;
   }

   if (curfle->flags & FILE_ISDIR) {
      ftp_log (gftp_logging_misc, NULL, "Edit: %s is a directory. Cannot edit it.\n", curfle->file);
      return;
   }
   
   if (wdata->local) view_file (curfle->file, 0, 0, 1, NULL);
   else {
      make_temp_filename (tempstr, strchr (curfle->file, '.'), sizeof (tempstr));
      newtdata = transfer_one_file (tempstr, curfle->file, 1);
      newtdata->hdata->files->flags |= FILE_TRANS_DONE_EDIT;
      add_file_transfer (newtdata);
   }
}
/*****************************************************************************/
void view_file (char *filename, int viewedit, int del_file, int start_pos, char *remote_filename) {
   GtkWidget *dialog, *view, *table, *tempwid;
   struct viewedit_data *newproc;
   char tempstr[MAXSTR], *fle;
   struct pix_ext *tempext;
   int stlen;
   FILE *fd;

   stlen = strlen (filename);
   tempext = registered_exts;
   while (tempext != NULL) {
      if (stlen >= tempext->stlen &&
         strcmp (&filename[stlen-tempext->stlen], tempext->ext) == 0) {

         if (*tempext->view_program == '\0') break;
         ftp_log (gftp_logging_misc, NULL, "Opening %s with %s\n", filename, tempext->view_program);
         fork_process (tempext->view_program, filename, remote_filename, viewedit, del_file);
         return;
      }
      tempext = tempext->next;
   }
   if (viewedit && *view_program != '\0') {
      /* Open the file with the default file viewer */
      fork_process (view_program, filename, remote_filename, viewedit, del_file);
      return;
   }
   else if (!viewedit) {
      /* Open the file with the default file editor */
      newproc = fork_process (edit_program, filename, remote_filename, viewedit, del_file);
      pthread_mutex_lock (&sigchld_mutex);
      stat (filename, &newproc->st);
      pthread_mutex_unlock (&sigchld_mutex);
      return;
   }

   if ((fd = fopen (filename, "r")) == NULL) {
      ftp_log (gftp_logging_misc, NULL, "View: Cannot open file %s: %s\n", filename, g_strerror (errno));
      return;
   }
   ftp_log (gftp_logging_misc, NULL, "Viewing file %s\n", filename);

   dialog = gtk_dialog_new ();
   gtk_window_set_title (GTK_WINDOW (dialog), filename);
   gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 5);
   gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (dialog)->action_area), 5);
   gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 5);
   gtk_box_set_homogeneous (GTK_BOX (GTK_DIALOG (dialog)->action_area), TRUE);
   gtk_signal_connect_object (GTK_OBJECT (dialog), "delete_event", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog));
   gtk_signal_connect_object (GTK_OBJECT (dialog), "destroy", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog));

   table = gtk_table_new (1, 2, FALSE);
   gtk_widget_set_usize (table, 500, 400);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), table, TRUE, TRUE, FALSE);

   view = gtk_text_new (NULL, NULL);
   gtk_text_set_editable (GTK_TEXT (view), FALSE);
   gtk_text_set_word_wrap (GTK_TEXT (view), TRUE);
   gtk_table_attach (GTK_TABLE (table), view, 0, 1, 0, 1,
   	GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 0);
   gtk_widget_show (view);

   tempwid = gtk_vscrollbar_new (GTK_TEXT (view)->vadj);
   gtk_table_attach (GTK_TABLE (table), tempwid, 1, 2, 0, 1,
   	GTK_FILL, GTK_EXPAND | GTK_FILL | GTK_SHRINK, 0, 0);
   gtk_widget_show (tempwid);
   gtk_widget_show (table);

   tempwid = gtk_button_new_with_label ("   Close   ");
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, FALSE, FALSE, FALSE);
   if (del_file) {
      fle = g_malloc (strlen (filename) + 1);
      strcpy (fle, filename);
      gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (delete_file), (gpointer) fle);
   }
   gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog));
   gtk_widget_show (tempwid);

   memset (tempstr, 0, sizeof (tempstr));
   while (fread (tempstr, 1, sizeof (tempstr) - 1, fd)) {
      gtk_text_insert (GTK_TEXT (view), NULL, NULL, NULL, tempstr, strlen (tempstr));
      memset (tempstr, 0, sizeof (tempstr));
   }
   fclose (fd);
   gtk_widget_show (dialog);
   if (!start_pos) {
      gtk_adjustment_set_value (GTK_TEXT(view)->vadj, 
      	GTK_TEXT (view)->vadj->upper - GTK_TEXT (view)->vadj->page_size);
   }
}
/*****************************************************************************/
static struct viewedit_data *fork_process (char *proc, char *filename, char *remote_filename, int viewedit, int del_file) {
   struct viewedit_data *newproc;
   pid_t ret;
   
   newproc = NULL;
   if ((ret = fork ()) == 0) {
      if (execlp (proc, proc, filename, NULL) == -1) {
         ftp_log (gftp_logging_misc, NULL, "Cannot run program %s: %s\n", proc, g_strerror (errno));
      }
   }
   else if (ret == -1) ftp_log (gftp_logging_misc, NULL, "View: Cannot fork another process: %s\n", g_strerror (errno));
   else {
      ftp_log (gftp_logging_misc, NULL, "Running program: %s %s\n", proc, filename);
      pthread_mutex_lock (&sigchld_mutex);
      newproc = g_malloc (sizeof (struct viewedit_data));
      newproc->next = viewedit_processes;
      newproc->done = 0;
      viewedit_processes = newproc;
      newproc->pid = ret;
      strncpy (newproc->filename, filename, sizeof (newproc->filename));
      newproc->filename[sizeof (newproc->filename) - 1] = '\0';
      if (remote_filename != NULL) {
         strncpy (newproc->remote_filename, remote_filename, sizeof (newproc->remote_filename));
         newproc->remote_filename[sizeof (newproc->remote_filename) - 1] = '\0';
      }
      newproc->flags = viewedit ? VIEWEDIT_VIEW_FILE : 0;
      if (del_file) newproc->flags |= VIEWEDIT_RM_FILE;
      pthread_mutex_unlock (&sigchld_mutex);
   }
   return (newproc);
}
/*****************************************************************************/
static void delete_file (GtkWidget *widget, char *fle) {
   unlink (fle);
   g_free (fle);
}
/*****************************************************************************/
